/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QAction>
#include <QApplication>
#include <QFont>
#include <QGridLayout>
#include <QScreen>
#include <QLabel>
#include <QSettings>
#include <QMenu>
#include <QMenuBar>
#include <QPlainTextEdit>
#include <QTextCursor>
#include <QWidget>
#include <QFile>
#include <QFileDialog>
#include <QTextStream>
#include <QByteArray>
#include <QSizePolicy>
#include <QFont>
#include <QDir>
#include <QRect>
#include <QSize>
#include <QPoint>

#include "validation2dmis.h"

TValidation2DMIS::TValidation2DMIS(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
	QFont								font;
	QGridLayout							*gridLayout;
	QLabel								*logo_label;
	QLabel								*copyright_label;
	QMenu								*file_menu;
	QMenu								*edit_menu;
	QMenuBar							*menubar;
	QWidget								*central_widget;
	QSizePolicy							sizepolicy_preferred_maximum(QSizePolicy::Preferred, QSizePolicy::Maximum);
	QRect								desktop_rect;
	QRect								window_rect;
	QPoint								position;
	QSize								size;
	QDir								data_path;
	int									font_size;
	QFont								copyright_font;
	int									ival;

	font.setFamily(QStringLiteral("Courier New"));
	
	copyright_font.setFamily(QStringLiteral("Helvetica"));
	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);

	font_size = font.pointSize();
	
	if(!(font_size < 0))
	{
		font.setPointSize(font_size + 2);
	}
	
#ifdef Q_OS_MAC
	int index = d_application_path.lastIndexOf(QStringLiteral(".app/"));
	d_application_path.truncate(index);
	
	index = d_application_path.lastIndexOf('/');
	d_application_path.truncate(index);
#endif
	
#ifdef Q_OS_WIN
	d_data_path = d_application_path;
#else
	d_data_path = QDir::homePath();
#endif
	
	d_file_path = d_data_path;
	
	data_path.setPath(d_data_path);
	
	if(!data_path.exists(QStringLiteral(".validation2dmis")))
	{
		data_path.mkdir(QStringLiteral(".validation2dmis"));
	}
	
	data_path.cd(QStringLiteral(".validation2dmis"));
	d_data_path = data_path.absolutePath();
	
	d_settings = new QSettings(d_data_path + QStringLiteral("/validation2dmis_settings.ini"),QSettings::IniFormat,this);
	
	d_file_path = d_settings->value("File_Path", d_file_path).toString();

	position = d_settings->value("Mainwindow_Position", QPoint(30,30)).toPoint();
	size = d_settings->value("Mainwindow_Size", QSize(800,600)).toSize();
	
	desktop_rect = QGuiApplication::primaryScreen()->availableGeometry();
	
	window_rect.moveTopLeft(position);
	window_rect.setSize(size);
	
	window_rect = window_rect & desktop_rect;
	
	if(window_rect.isValid() && window_rect.width() > 500 && window_rect.height() > 500)
	{
		this->resize(window_rect.size());
		this->move(window_rect.topLeft());
	}
	else
	{
		this->resize(800,600);
		this->move(QPoint(30,30));
	}
	
	ival = d_settings->value("DMIS_Feature_Type", 0).toInt();
	d_preferences.dmis_output_feature = static_cast<TPreferencesDialog::TDMISOutputFeature>(ival);

	d_action_open = new QAction(this);
	
	d_action_export_dmis = new QAction(this);
	
	d_action_quit = new QAction(this);
	d_action_quit->setShortcut(QKeySequence::Quit);

	d_action_preferences = new QAction(this);

	central_widget = new QWidget(this);
	this->setCentralWidget(central_widget);

	gridLayout = new QGridLayout(central_widget);

	d_text_edit = new QPlainTextEdit(central_widget);
	d_text_edit->setFont(font);
	d_text_edit->setLineWrapMode(QPlainTextEdit::NoWrap);
	d_text_edit->setReadOnly(true);
	gridLayout->addWidget(d_text_edit,0,0,1,2);
	
	
	copyright_label = new QLabel(central_widget);
	copyright_label->setFont(copyright_font);
	sizepolicy_preferred_maximum.setHeightForWidth(copyright_label->sizePolicy().hasHeightForWidth());
	copyright_label->setSizePolicy(sizepolicy_preferred_maximum);
	copyright_label->setAlignment(Qt::AlignBottom | Qt::AlignLeft);	//Qt::AlignRight|Qt::AlignTrailing
	copyright_label->setMinimumHeight(40);
	gridLayout->addWidget(copyright_label,1,0,1,1);

	logo_label = new QLabel(central_widget);
	logo_label->setMinimumSize(100,40);
	logo_label->setMaximumSize(100,40);
	gridLayout->addWidget(logo_label,1,1,1,1);

	menubar = new QMenuBar(this);
	this->setMenuBar(menubar);
	
	file_menu = new QMenu(menubar);
	menubar->addMenu(file_menu);
	
	edit_menu = new QMenu(menubar);
	menubar->addMenu(edit_menu);
	
	file_menu->addAction(d_action_open);
	file_menu->addAction(d_action_export_dmis);
	file_menu->addAction(d_action_quit);

	edit_menu->addAction(d_action_preferences);

	menubar->addAction(file_menu->menuAction());
	menubar->addAction(edit_menu->menuAction());

	// defaults
	d_preferences_dialog = 0;
	d_action_export_dmis->setEnabled(false);
	
	logo_label->setPixmap(QPixmap(QStringLiteral(":/images/logo.png")));

	this->setWindowTitle(QStringLiteral("Validation Data To DMIS - Version 1.0"));
	
	file_menu->setTitle(QStringLiteral("File"));
	edit_menu->setTitle(QStringLiteral("Edit"));
	d_action_open->setText(QStringLiteral("Open"));
	d_action_export_dmis->setText(QStringLiteral("Export DMIS"));
	d_action_quit->setText(QStringLiteral("Quit"));
	d_action_preferences->setText(QStringLiteral("Preferences"));
	copyright_label->setText(QStringLiteral("Copyright (C) 2023 Select Calibration Incorporated.  www.selectcalibration.ca"));
	
	this->Display_Program_Information();

	connect(d_action_open,&QAction::triggered,this,&TValidation2DMIS::Menu_Open);
	connect(d_action_export_dmis,&QAction::triggered,this,&TValidation2DMIS::Menu_Export_DMIS);
	connect(d_action_quit,&QAction::triggered,this,&TValidation2DMIS::close);
	connect(d_action_preferences,&QAction::triggered,this,&TValidation2DMIS::Menu_Preferences);
}

TValidation2DMIS::~TValidation2DMIS(void)
{
}

void TValidation2DMIS::Menu_Open(void)
{
	QString								file_name;
	QByteArray							file_data;
	QFile								file;
	int									index;
	
	file_name = QFileDialog::getOpenFileName(this,
											 QStringLiteral("Open Validation File"),
											 d_file_path,
											 QStringLiteral("Validation File (*.dat)"));
	
	if(file_name.length())
	{
		file.setFileName(file_name);
		
		if(file.open(QIODevice::ReadOnly))
		{
			file_data = file.readAll();
			file.close();
			
			this->Clear_Text();
			
			this->Create_Header();
			this->Process_Validation_File(file_data);
			this->Create_Footer();
			
			this->Update_Display();
			this->Scroll_Text_Top();
			
			d_action_export_dmis->setEnabled(true);
		}
		
		index = file_name.lastIndexOf('/');
		
		if(!(index < 0))
		{
			d_file_path = file_name;
			d_file_path.truncate(index);
		}
	}
}

void TValidation2DMIS::Menu_Export_DMIS(void)
{
	QString								file_name;
	QByteArray							file_data;
	QFile								file;
	QString								text;
	int									index;

	file_name = QFileDialog::getSaveFileName(this,
											 QStringLiteral("Create DMIS File"),
											 d_file_path,
											 QStringLiteral("DMIS (*.dms)"));
	
	if(file_name.length())
	{
		file.setFileName(file_name);
		
		if(file.open(QIODevice::WriteOnly))
		{
			text = d_text_edit->toPlainText();
			file.write(text.toLatin1());
			file.close();
		}
		
		index = file_name.lastIndexOf('/');
		
		if(!(index < 0))
		{
			d_file_path = file_name;
			d_file_path.truncate(index);
		}
	}
}

void TValidation2DMIS::Menu_Preferences(void)
{
	TPreferencesDialog::TPreferences	preferences;
	
	if(!d_preferences_dialog)
	{
		d_preferences_dialog = new TPreferencesDialog(this);
	}
	
	d_preferences_dialog->Reset(d_preferences);
	
	d_preferences_dialog->exec();
	
	preferences = d_preferences_dialog->Preferences();
	
	if(preferences.dmis_output_feature != d_preferences.dmis_output_feature)
	{
		this->Clear_Text();
	}
	
	d_preferences = preferences;
}

void TValidation2DMIS::closeEvent(
	QCloseEvent							*event)
{
	d_settings->setValue("Mainwindow_Position", this->pos());
	d_settings->setValue("Mainwindow_Size", this->size());
	d_settings->setValue("DMIS_Feature_Type", static_cast<int>(d_preferences.dmis_output_feature));
	d_settings->setValue("File_Path", d_file_path);

	event->accept();
}

void TValidation2DMIS::Clear_Text(void)
{
	d_text_edit->clear();
	d_construction_features.clear();
	d_dmis_text_buffer.clear();
	
	d_action_export_dmis->setEnabled(false);
}

void TValidation2DMIS::Add_Text(
	const QString						&text)
{
	d_dmis_text_buffer.push_back(text);	
	d_dmis_text_buffer.push_back('\n');
}

void TValidation2DMIS::Update_Display(void)
{
	d_text_edit->moveCursor(QTextCursor::End);
	d_text_edit->appendPlainText(d_dmis_text_buffer);
}

void TValidation2DMIS::Scroll_Text_Top(void)
{
	d_text_edit->moveCursor(QTextCursor::Start);
}

void TValidation2DMIS::Create_Header(void)
{

	this->Add_Text(QStringLiteral("$$ Validation DMIS file"));
	this->Add_Text(QStringLiteral("DMISMN/'validation.dms' 04.0"));
	this->Add_Text(QStringLiteral("FILNAM/'validation.dms'"));
	this->Add_Text(QStringLiteral("$$"));
	this->Add_Text(QStringLiteral("UNITS/MM,ANGDEC"));
	this->Add_Text(QStringLiteral("D(MC_CSY)=DATSET/MCS"));
	this->Add_Text(QStringLiteral("MODE/MAN"));
	this->Add_Text(QStringLiteral("$$"));
	this->Add_Text(QStringLiteral("$$ Increase the display precision to 5 decimal places minimum."));
	this->Add_Text(QStringLiteral("DECPL/ALL,5"));
	
	if(d_preferences.dmis_output_feature == TPreferencesDialog::OFFSET_POINT)
	{
		this->Add_Text(QStringLiteral("$$"));
		this->Add_Text(QStringLiteral("F(PT_ZERO)=FEAT/POINT,0,0,0,0,0,1"));
	}

}

void TValidation2DMIS::Process_Validation_File(
	const QByteArray					&file_data)
{
	QTextStream							stream(file_data,QIODevice::ReadOnly);
	QString								line;
	QString								text;
	QString								text_section;
	QStringList							list;
	bool								init(true);
	TVector3							position;
	TVector3							vector;
	double								form;
	double								diameter;
	int									point_count(0);
	int									feature_count(0);
	int									point_cntr(0);
	TValidation2DMIS::TFeatureType		feature_type(TValidation2DMIS::FEATURE_UNKNOWN);
	TValidation2DMIS::TConstructionFeatures construction_features;
	
	while(!stream.atEnd())
	{
		line = stream.readLine().simplified();
		
		if(line.startsWith('#'))
		{
			continue;
		}
		
		if(init)
		{
			if(line.startsWith(QStringLiteral("Best_Fit_Validation:Version=1:Type=Plane")))
			{
				feature_type = TValidation2DMIS::FEATURE_PLANE;
			}
			else if(line.startsWith(QStringLiteral("Best_Fit_Validation:Version=1:Type=Line")))
			{
				feature_type = TValidation2DMIS::FEATURE_LINE;
			}
			else if(line.startsWith(QStringLiteral("Best_Fit_Validation:Version=1:Type=Circle")))
			{
				feature_type = TValidation2DMIS::FEATURE_CIRCLE;
			}
			else if(line.startsWith(QStringLiteral("Best_Fit_Validation:Version=1:Type=Sphere")))
			{
				feature_type = TValidation2DMIS::FEATURE_SPHERE;
			}
			else
			{
				// unknown file type
				return;
			}
			
			point_count = 0;
			feature_count = 0;
			
			init = false;
		}
		else
		{
			if(feature_type == TValidation2DMIS::FEATURE_PLANE)
			{
				if(line.startsWith(QStringLiteral("Plane"),Qt::CaseInsensitive))
				{
					text = line.mid(6);	// remove 'Plane '
					list = text.split(' ');
					
					if(list.size() == 9)
					{
						feature_count = list[0].toInt();
						point_count = list[1].toInt();
						
						point_cntr = 1;
						
						position.x = list[2].toDouble();
						position.y = list[3].toDouble();
						position.z = list[4].toDouble();
						
						vector.x = list[5].toDouble();
						vector.y = list[6].toDouble();
						vector.z = list[7].toDouble();
						
						form = list[8].toDouble();
						
						diameter = 0.0;
						
						this->Add_Text(QStringLiteral("$$"));

						text = QString("$$ Plane %1 %2 %3 %4 %5 %6 %7")
							.arg(position.x,0,'f',6)
							.arg(position.y,0,'f',6)
							.arg(position.z,0,'f',6)
							.arg(vector.i,0,'f',9)
							.arg(vector.j,0,'f',9)
							.arg(vector.k,0,'f',9)
							.arg(form,0,'f',6);
						
						this->Add_Text(text);
						
						construction_features.feature_name = QString("PLN_%1").arg(feature_count);
						construction_features.position = position;
						construction_features.vector = vector;
						construction_features.diameter = 0.0;
						
						construction_features.points.clear();
					}
				}
				else if(point_count > 0)
				{
					list = line.split(' ');
					
					if(list.size() == 3)
					{
						position.x = list[0].toDouble();
						position.y = list[1].toDouble();
						position.z = list[2].toDouble();
						
						vector.Set(0,0,1);
						
						this->Add_Text(QStringLiteral("$$"));
						
						text = QString("F(PT_%1_%2)=FEAT/POINT,CART,%3,%4,%5,0.000000000,0.000000000,1.000000000")
								.arg(feature_count)
								.arg(point_cntr)
								.arg(position.x,0,'f',6)
								.arg(position.y,0,'f',6)
								.arg(position.z,0,'f',6);
						this->Add_Text(text);

						switch(d_preferences.dmis_output_feature)
						{
							case TPreferencesDialog::MEASURED_POINT:
								text = QString("MEAS/POINT,F(PT_%1_%2),1")
											.arg(feature_count)
											.arg(point_cntr);
								this->Add_Text(text);
								
								text = QString("    PTMEAS/CART,%1,%2,%3,0.000000000,0.000000000,1.000000000")
											.arg(position.x,0,'f',6)
											.arg(position.y,0,'f',6)
											.arg(position.z,0,'f',6);
								this->Add_Text(text);
								
								this->Add_Text(QStringLiteral("ENDMES"));
								break;
								
							case TPreferencesDialog::OFFSET_POINT:
								text = QString("CONST/POINT,F(PT_%1_%2),MOVEPT,F(PT_ZERO),%3,%4,%5")
											.arg(feature_count)
											.arg(point_cntr)
											.arg(position.x,0,'f',6)
											.arg(position.y,0,'f',6)
											.arg(position.z,0,'f',6);
								this->Add_Text(text);
								break;
						}
						
						text = QString("PT_%1_%2")
								.arg(feature_count)
								.arg(point_cntr);

						construction_features.points.push_back(text);
						
						point_count--;
						point_cntr++;
						
						if(point_count == 0)
						{
							d_construction_features.push_back(construction_features);
						}
					}
				}
				else
				{
					this->Add_Text(QStringLiteral("$$  ERR:  Number of data points greater than defined in feature."));
				}
			}
			else if(feature_type == TValidation2DMIS::FEATURE_LINE)
			{
				if(line.startsWith(QStringLiteral("Line"),Qt::CaseInsensitive))
				{
					text = line.mid(5);	// remove 'Line '
					list = text.split(' ');
					
					if(list.size() == 9)
					{
						feature_count = list[0].toInt();
						point_count = list[1].toInt();
						
						point_cntr = 1;
						
						position.x = list[2].toDouble();
						position.y = list[3].toDouble();
						position.z = list[4].toDouble();
						
						vector.x = list[5].toDouble();
						vector.y = list[6].toDouble();
						vector.z = list[7].toDouble();
						
						form = list[8].toDouble();
						
						diameter = 0.0;
						
						this->Add_Text(QStringLiteral("$$"));

						text = QString("$$ Line %1 %2 %3 %4 %5 %6 %7")
							.arg(position.x,0,'f',6)
							.arg(position.y,0,'f',6)
							.arg(position.z,0,'f',6)
							.arg(vector.i,0,'f',9)
							.arg(vector.j,0,'f',9)
							.arg(vector.k,0,'f',9)
							.arg(form,0,'f',6);
						
						this->Add_Text(text);
						
						construction_features.feature_name = QString("LN_%1").arg(feature_count);
						construction_features.position = position;
						construction_features.vector = vector;
						construction_features.diameter = 0.0;
						
						construction_features.points.clear();
					}
				}
				else if(point_count > 0)
				{
					list = line.split(' ');
					
					if(list.size() == 3)
					{
						position.x = list[0].toDouble();
						position.y = list[1].toDouble();
						position.z = list[2].toDouble();
						
						vector.Set(0,0,1);
						
						this->Add_Text(QStringLiteral("$$"));
						
						text = QString("F(PT_%1_%2)=FEAT/POINT,CART,%3,%4,%5,0.000000000,0.000000000,1.000000000")
								.arg(feature_count)
								.arg(point_cntr)
								.arg(position.x,0,'f',6)
								.arg(position.y,0,'f',6)
								.arg(position.z,0,'f',6);
						this->Add_Text(text);
						
						switch(d_preferences.dmis_output_feature)
						{
							case TPreferencesDialog::MEASURED_POINT:
								text = QString("MEAS/POINT,F(PT_%1_%2),1")
											.arg(feature_count)
											.arg(point_cntr);
								this->Add_Text(text);
								
								text = QString("    PTMEAS/CART,%1,%2,%3,0.000000000,0.000000000,1.000000000")
											.arg(position.x,0,'f',6)
											.arg(position.y,0,'f',6)
											.arg(position.z,0,'f',6);
								this->Add_Text(text);
								
								this->Add_Text(QStringLiteral("ENDMES"));
								break;
								
							case TPreferencesDialog::OFFSET_POINT:
								text = QString("CONST/POINT,F(PT_%1_%2),MOVEPT,F(PT_ZERO),%3,%4,%5")
											.arg(feature_count)
											.arg(point_cntr)
											.arg(position.x,0,'f',6)
											.arg(position.y,0,'f',6)
											.arg(position.z,0,'f',6);
								this->Add_Text(text);
								break;
						}

						text = QString("PT_%1_%2")
								.arg(feature_count)
								.arg(point_cntr);

						construction_features.points.push_back(text);
						
						point_count--;
						point_cntr++;
						
						if(point_count == 0)
						{
							d_construction_features.push_back(construction_features);
						}
					}
				}
				else
				{
					this->Add_Text(QStringLiteral("$$  ERR:  Number of data points greater than defined in feature."));
				}
			}
			else if(feature_type == TValidation2DMIS::FEATURE_CIRCLE)
			{
				if(line.startsWith(QStringLiteral("Circle"),Qt::CaseInsensitive))
				{
					text = line.mid(7);	// remove 'Circle '
					list = text.split(' ');
					
					if(list.size() == 10)
					{
						feature_count = list[0].toInt();
						point_count = list[1].toInt();
						
						point_cntr = 1;
						
						position.x = list[2].toDouble();
						position.y = list[3].toDouble();
						position.z = list[4].toDouble();
						
						vector.x = list[5].toDouble();
						vector.y = list[6].toDouble();
						vector.z = list[7].toDouble();
						
						diameter = list[8].toDouble();
						form = list[9].toDouble();
												
						this->Add_Text(QStringLiteral("$$"));

						text = QString("$$ Circle %1 %2 %3 %4 %5 %6 %7 %8")
							.arg(position.x,0,'f',6)
							.arg(position.y,0,'f',6)
							.arg(position.z,0,'f',6)
							.arg(vector.i,0,'f',9)
							.arg(vector.j,0,'f',9)
							.arg(vector.k,0,'f',9)
							.arg(diameter,0,'f',6)
							.arg(form,0,'f',6);
						
						this->Add_Text(text);
						
						construction_features.feature_name = QString("CIR_%1").arg(feature_count);
						construction_features.position = position;
						construction_features.vector = vector;
						construction_features.diameter = diameter;
						
						construction_features.points.clear();
					}
				}
				else if(point_count > 0)
				{
					list = line.split(' ');
					
					if(list.size() == 3)
					{
						position.x = list[0].toDouble();
						position.y = list[1].toDouble();
						position.z = list[2].toDouble();
						
						vector.Set(0,0,1);
						
						this->Add_Text(QStringLiteral("$$"));
						
						text = QString("F(PT_%1_%2)=FEAT/POINT,CART,%3,%4,%5,0.000000000,0.000000000,1.000000000")
								.arg(feature_count)
								.arg(point_cntr)
								.arg(position.x,0,'f',6)
								.arg(position.y,0,'f',6)
								.arg(position.z,0,'f',6);
						this->Add_Text(text);
						
						switch(d_preferences.dmis_output_feature)
						{
							case TPreferencesDialog::MEASURED_POINT:
								text = QString("MEAS/POINT,F(PT_%1_%2),1")
											.arg(feature_count)
											.arg(point_cntr);
								this->Add_Text(text);
								
								text = QString("    PTMEAS/CART,%1,%2,%3,0.000000000,0.000000000,1.000000000")
											.arg(position.x,0,'f',6)
											.arg(position.y,0,'f',6)
											.arg(position.z,0,'f',6);
								this->Add_Text(text);
								
								this->Add_Text(QStringLiteral("ENDMES"));
								break;
								
							case TPreferencesDialog::OFFSET_POINT:
								text = QString("CONST/POINT,F(PT_%1_%2),MOVEPT,F(PT_ZERO),%3,%4,%5")
											.arg(feature_count)
											.arg(point_cntr)
											.arg(position.x,0,'f',6)
											.arg(position.y,0,'f',6)
											.arg(position.z,0,'f',6);
								this->Add_Text(text);
								break;
						}

						text = QString("PT_%1_%2")
								.arg(feature_count)
								.arg(point_cntr);

						construction_features.points.push_back(text);
						
						point_count--;
						point_cntr++;
						
						if(point_count == 0)
						{
							d_construction_features.push_back(construction_features);
						}
					}
				}
				else
				{
					this->Add_Text(QStringLiteral("$$  ERR:  Number of data points greater than defined in feature."));
				}
			}
			else if(feature_type == TValidation2DMIS::FEATURE_SPHERE)
			{
				if(line.startsWith(QStringLiteral("Sphere"),Qt::CaseInsensitive))
				{
					text = line.mid(7);	// remove 'Sphere '
					list = text.split(' ');
					
					if(list.size() == 7)
					{
						feature_count = list[0].toInt();
						point_count = list[1].toInt();
						
						point_cntr = 1;
						
						position.x = list[2].toDouble();
						position.y = list[3].toDouble();
						position.z = list[4].toDouble();
												
						diameter = list[5].toDouble();
						form = list[6].toDouble();
												
						this->Add_Text(QStringLiteral("$$"));

						text = QString("$$ Sphere %1 %2 %3 %4 %5")
							.arg(position.x,0,'f',6)
							.arg(position.y,0,'f',6)
							.arg(position.z,0,'f',6)
							.arg(diameter,0,'f',6)
							.arg(form,0,'f',6);
						
						this->Add_Text(text);
						
						construction_features.feature_name = QString("SPH_%1").arg(feature_count);
						construction_features.position = position;
						construction_features.vector = vector;
						construction_features.diameter = diameter;
						
						construction_features.points.clear();
					}
				}
				else if(point_count > 0)
				{
					list = line.split(' ');
					
					if(list.size() == 3)
					{
						position.x = list[0].toDouble();
						position.y = list[1].toDouble();
						position.z = list[2].toDouble();
						
						vector.Set(0,0,1);
						
						this->Add_Text(QStringLiteral("$$"));
						
						text = QString("F(PT_%1_%2)=FEAT/POINT,CART,%3,%4,%5,0.000000000,0.000000000,1.000000000")
								.arg(feature_count)
								.arg(point_cntr)
								.arg(position.x,0,'f',6)
								.arg(position.y,0,'f',6)
								.arg(position.z,0,'f',6);
						this->Add_Text(text);
						
						switch(d_preferences.dmis_output_feature)
						{
							case TPreferencesDialog::MEASURED_POINT:
								text = QString("MEAS/POINT,F(PT_%1_%2),1")
											.arg(feature_count)
											.arg(point_cntr);
								this->Add_Text(text);
								
								text = QString("    PTMEAS/CART,%1,%2,%3,0.000000000,0.000000000,1.000000000")
											.arg(position.x,0,'f',6)
											.arg(position.y,0,'f',6)
											.arg(position.z,0,'f',6);
								this->Add_Text(text);
								
								this->Add_Text(QStringLiteral("ENDMES"));
								break;
								
							case TPreferencesDialog::OFFSET_POINT:
								text = QString("CONST/POINT,F(PT_%1_%2),MOVEPT,F(PT_ZERO),%3,%4,%5")
											.arg(feature_count)
											.arg(point_cntr)
											.arg(position.x,0,'f',6)
											.arg(position.y,0,'f',6)
											.arg(position.z,0,'f',6);
								this->Add_Text(text);
								break;
						}

						text = QString("PT_%1_%2")
								.arg(feature_count)
								.arg(point_cntr);

						construction_features.points.push_back(text);
						
						point_count--;
						point_cntr++;
						
						if(point_count == 0)
						{
							d_construction_features.push_back(construction_features);
						}
					}
				}
				else
				{
					this->Add_Text(QStringLiteral("$$  ERR:  Number of data points greater than defined in feature."));
				}
			}
		}
	}
	
	switch(feature_type)
	{
		case TValidation2DMIS::FEATURE_UNKNOWN:
			break;

		case TValidation2DMIS::FEATURE_PLANE:
			this->Create_Constructed_Planes();
			break;
			
		case TValidation2DMIS::FEATURE_LINE:
			this->Create_Constructed_Lines();
			break;
			
		case TValidation2DMIS::FEATURE_CIRCLE:
			this->Create_Constructed_Circles();
			break;
			
		case TValidation2DMIS::FEATURE_SPHERE:
			this->Create_Constructed_Spheres();
			break;
	}
}

void TValidation2DMIS::Create_Footer(void)
{
	this->Add_Text(QStringLiteral("ENDFIL"));
}

void TValidation2DMIS::Display_Program_Information(void)
{
	this->Add_Text(QStringLiteral(" "));
	this->Add_Text(QStringLiteral(" "));
	this->Add_Text(QStringLiteral(" "));
	this->Add_Text(QStringLiteral("Validation Data to DMIS Converter"));
	this->Add_Text(QStringLiteral("======================================================="));
	this->Add_Text(QStringLiteral(" "));
	this->Add_Text(QStringLiteral("Version: 1.0"));
	this->Add_Text(QStringLiteral(" "));
	this->Add_Text(QStringLiteral("This utility will generate a DMIS program from the validation data"));
	this->Add_Text(QStringLiteral("generated by MeasureDirect.  The DMIS program can be imported into"));
	this->Add_Text(QStringLiteral("inspection software for comparison purposes."));
	
	this->Update_Display();
}

void TValidation2DMIS::Create_Constructed_Planes(void)
{
	std::vector<TValidation2DMIS::TConstructionFeatures>::const_iterator feature_iter;
	std::vector<QString>::const_iterator point_iter;
	QString								text;
	int									cntr;
	static const int					INIT_FEATURE_COUNT(2);
	static const int					MAX_FEATURE_COUNT(6);
	
	for(feature_iter = d_construction_features.begin();feature_iter != d_construction_features.end();++feature_iter)
	{
		this->Add_Text(QStringLiteral("$$"));

		text = QString("F(%1)=FEAT/PLANE,CART,%2,%3,%4,%5,%6,%7")
					.arg((*feature_iter).feature_name)
					.arg((*feature_iter).position.x,0,'f',6)
					.arg((*feature_iter).position.y,0,'f',6)
					.arg((*feature_iter).position.z,0,'f',6)
					.arg((*feature_iter).vector.i,0,'f',9)
					.arg((*feature_iter).vector.j,0,'f',9)
					.arg((*feature_iter).vector.k,0,'f',9);
		
		this->Add_Text(text);

		text = QString("CONST/PLANE,F(%1),BF").arg((*feature_iter).feature_name);
		
		cntr = INIT_FEATURE_COUNT;
		
		for(point_iter = (*feature_iter).points.begin();point_iter != (*feature_iter).points.end();++point_iter)
		{
			if(cntr > MAX_FEATURE_COUNT)
			{
				text += '$';
				
				this->Add_Text(text);
				text.clear();
				
				cntr = 0;
			}
			
			text += QString(",FA(%1)").arg(*point_iter);
			cntr++;
		}
		
		if(cntr)
		{
			this->Add_Text(text);
			text.clear();
		}
	}
}

void TValidation2DMIS::Create_Constructed_Lines(void)
{
	std::vector<TValidation2DMIS::TConstructionFeatures>::const_iterator feature_iter;
	std::vector<QString>::const_iterator point_iter;
	QString								text;
	int									cntr;
	static const int					INIT_FEATURE_COUNT(2);
	static const int					MAX_FEATURE_COUNT(6);
	
	for(feature_iter = d_construction_features.begin();feature_iter != d_construction_features.end();++feature_iter)
	{
		this->Add_Text(QStringLiteral("$$"));

		text = QString("F(%1)=FEAT/LINE,UNBND,CART,%2,%3,%4,%5,%6,%7,0.000000000,0.000000000,1.000000000")
					.arg((*feature_iter).feature_name)
					.arg((*feature_iter).position.x,0,'f',6)
					.arg((*feature_iter).position.y,0,'f',6)
					.arg((*feature_iter).position.z,0,'f',6)
					.arg((*feature_iter).vector.i,0,'f',9)
					.arg((*feature_iter).vector.j,0,'f',9)
					.arg((*feature_iter).vector.k,0,'f',9);
		
		this->Add_Text(text);

		text = QString("CONST/LINE,F(%1),BF").arg((*feature_iter).feature_name);
		
		cntr = INIT_FEATURE_COUNT;
		
		for(point_iter = (*feature_iter).points.begin();point_iter != (*feature_iter).points.end();++point_iter)
		{
			if(cntr > MAX_FEATURE_COUNT)
			{
				text += '$';
				
				this->Add_Text(text);
				text.clear();
				
				cntr = 0;
			}
			
			text += QString(",FA(%1)").arg(*point_iter);
			cntr++;
		}
		
		if(cntr)
		{
			this->Add_Text(text);
			text.clear();
		}
	}
}

void TValidation2DMIS::Create_Constructed_Circles(void)
{
	std::vector<TValidation2DMIS::TConstructionFeatures>::const_iterator feature_iter;
	std::vector<QString>::const_iterator point_iter;
	QString								text;
	int									cntr;
	static const int					INIT_FEATURE_COUNT(2);
	static const int					MAX_FEATURE_COUNT(6);
	
	for(feature_iter = d_construction_features.begin();feature_iter != d_construction_features.end();++feature_iter)
	{
		this->Add_Text(QStringLiteral("$$"));

		text = QString("F(%1)=FEAT/CIRCLE,INNER,CART,%2,%3,%4,%5,%6,%7,%8")
					.arg((*feature_iter).feature_name)
					.arg((*feature_iter).position.x,0,'f',6)
					.arg((*feature_iter).position.y,0,'f',6)
					.arg((*feature_iter).position.z,0,'f',6)
					.arg((*feature_iter).vector.i,0,'f',9)
					.arg((*feature_iter).vector.j,0,'f',9)
					.arg((*feature_iter).vector.k,0,'f',9)
					.arg((*feature_iter).diameter,0,'f',6);
		
		this->Add_Text(text);

		text = QString("CONST/CIRCLE,F(%1),BF").arg((*feature_iter).feature_name);
		
		cntr = INIT_FEATURE_COUNT;
		
		for(point_iter = (*feature_iter).points.begin();point_iter != (*feature_iter).points.end();++point_iter)
		{
			if(cntr > MAX_FEATURE_COUNT)
			{
				text += '$';
				
				this->Add_Text(text);
				text.clear();
				
				cntr = 0;
			}
			
			text += QString(",FA(%1)").arg(*point_iter);
			cntr++;
		}
		
		if(cntr)
		{
			this->Add_Text(text);
			text.clear();
		}
	}
}

void TValidation2DMIS::Create_Constructed_Spheres(void)
{
	std::vector<TValidation2DMIS::TConstructionFeatures>::const_iterator feature_iter;
	std::vector<QString>::const_iterator point_iter;
	QString								text;
	int									cntr;
	static const int					INIT_FEATURE_COUNT(2);
	static const int					MAX_FEATURE_COUNT(6);
	
	for(feature_iter = d_construction_features.begin();feature_iter != d_construction_features.end();++feature_iter)
	{
		this->Add_Text(QStringLiteral("$$"));

		text = QString("F(%1)=FEAT/SPHERE,OUTER,CART,%2,%3,%4,%5")
					.arg((*feature_iter).feature_name)
					.arg((*feature_iter).position.x,0,'f',6)
					.arg((*feature_iter).position.y,0,'f',6)
					.arg((*feature_iter).position.z,0,'f',6)
					.arg((*feature_iter).diameter,0,'f',6);
		
		this->Add_Text(text);

		text = QString("CONST/SPHERE,F(%1),BF").arg((*feature_iter).feature_name);
		
		cntr = INIT_FEATURE_COUNT;
		
		for(point_iter = (*feature_iter).points.begin();point_iter != (*feature_iter).points.end();++point_iter)
		{
			if(cntr > MAX_FEATURE_COUNT)
			{
				text += '$';
				
				this->Add_Text(text);
				text.clear();
				
				cntr = 0;
			}
			
			text += QString(",FA(%1)").arg(*point_iter);
			cntr++;
		}
		
		if(cntr)
		{
			this->Add_Text(text);
			text.clear();
		}
	}
}
